'use strict';

var chalk = require('chalk');
var commander = require('commander');
var fs = require('fs-extra');
var path = require('path');
var execSync = require('child_process').execSync;
var spawn = require('cross-spawn');
var semver = require('semver');
var packageJson = require('./package.json');

var projectName = undefined;

var program = new commander.Command(packageJson.name).version(packageJson.version).arguments('<project-directory>').action(function (name) {
  projectName = name;
}).parse(process.argv);

if (typeof projectName === 'undefined') {
  console.log();
  console.log();
  console.log('Please enter the project directory:  ' + chalk.green(program.name()) + ' ' + chalk.red('<project-directory>'));
  console.log();
  console.log();
  process.exit(1);
}

createApp(projectName, program.verbose, program.scriptsVersion);

function createApp(name, verbose, version) {
  var root = path.resolve(name);
  var appName = path.basename(root);

  fs.ensureDirSync(name);
  if (!isSafeToCreateProjectIn(root)) {
    console.log('The directory ' + chalk.cyan(name) + ' contains files that could conflict.');
    console.log('Try using a new directory name.');
    process.exit(1);
  }

  console.log('Creating a new project in ' + chalk.cyan(root) + '.');
  console.log();

  var packageJson = {
    name: appName,
    version: '0.1.0',
    'private': true
  };

  fs.writeFileSync(path.join(root, 'package.json'), JSON.stringify(packageJson, null, 2));

  var originalDirectory = process.cwd();
  process.chdir(root);

  run(root, appName, version, verbose, originalDirectory);
}

function run(root, appName, version, verbose, originalDirectory) {
  var packageToInstall = getInstallPackage(version);

  console.log('Installing template...');

  var useYarn = shouldUseYarn();
  getPackageName(packageToInstall).then(function (packageName) {

    console.log('Installing ' + chalk.cyan(packageName) + '...');
    console.log();

    // Install template
    return install(useYarn, [packageToInstall], verbose).then(function () {
      return packageName;
    });
  }).then(function (packageName) {

    // Get template path
    var scriptsPath = path.resolve(process.cwd(), 'node_modules', packageName, 'template');

    // Copy template file to root directory
    fs.copySync(scriptsPath, root);

    console.log('Installing packages. This might take a couple minutes.');
    // Install all package from the template
    installProject(useYarn).then(function () {
      printSuccessLog(appName, root, useYarn);
    });
  });
}

function installProject(useYarn) {
  return new Promise(function (resolve, reject) {
    var command = undefined;
    var args = undefined;

    if (useYarn) {
      command = 'yarnpkg';
    } else {
      command = 'npm';
      args = ['install'];
    }

    var child = spawn(command, args, { stdio: 'inherit' });
    child.on('close', function (code) {
      if (code !== 0) {
        reject({
          command: command + ' ' + args.join(' ')
        });
        return;
      }
      resolve();
    });
  });
}

function install(useYarn, dependencies, verbose, isOnline) {
  return new Promise(function (resolve, reject) {
    var command = undefined;
    var args = undefined;
    if (useYarn) {
      command = 'yarnpkg';
      args = ['add', '--exact'];
      [].push.apply(args, dependencies);
    } else {
      command = 'npm';
      args = ['install', '--save', '--save-exact'].concat(dependencies);
    }

    if (verbose) {
      args.push('--verbose');
    }

    var child = spawn(command, args, { stdio: 'inherit' });
    child.on('close', function (code) {
      if (code !== 0) {
        reject({
          command: command + ' ' + args.join(' ')
        });
        return;
      }
      resolve();
    });
  });
}

function printSuccessLog(appName, appPath, useYarn) {

  var displayedCommand = useYarn ? 'yarn' : 'npm run';

  console.log();
  console.log();
  console.log(chalk.green('Success! Created project ' + chalk.cyan(appName)));
  console.log();
  console.log();
  console.log('You can run these commands in the project:');
  console.log('-------------------------------------------');
  console.log('| ' + chalk.cyan('cd ' + appName) + ' : Go to project directory.');
  console.log('| ' + chalk.cyan(displayedCommand + ' dev') + ' : Starts the development server.');
  console.log('| ' + chalk.cyan(displayedCommand + ' build:client') + ' : (Client side) Bundles files for production.');
  console.log('| ' + chalk.cyan(displayedCommand + ' build') + ' : (Server side) Bundles files for production.');
  console.log('| ' + chalk.cyan(displayedCommand + ' start') + ' : Starts the production server.');
  console.log('| ' + chalk.cyan(displayedCommand + ' test') + ' : Starts the test runner.');
  console.log('-------------------------------------------');
  console.log();
  console.log();
}

// Extract package name from tarball url or path.
function getPackageName(installPackage) {
  return Promise.resolve(installPackage);
}

function shouldUseYarn() {
  try {
    execSync('yarnpkg --version', { stdio: 'ignore' });
    return true;
  } catch (e) {
    return false;
  }
}

function getInstallPackage(version) {
  var packageToInstall = 'react-redux-template-script';
  var validSemver = semver.valid(version);
  if (validSemver) {
    packageToInstall += '@' + validSemver;
  } else if (version) {
    packageToInstall = version;
  }
  return packageToInstall;
}

function isSafeToCreateProjectIn(root) {
  var validFiles = ['.DS_Store', 'Thumbs.db', '.git', '.gitignore', '.idea', 'README.md', 'LICENSE', 'web.iml', '.hg', '.hgignore', '.hgcheck'];
  return fs.readdirSync(root).every(function (file) {
    return validFiles.indexOf(file) >= 0;
  });
}